"""."""

import numpy as np
import pytest

from cv2 import KalmanFilter

from kinematic_tracker.nd.gen_kf import NdKkfGenKf
from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz
from kinematic_tracker.nd.shape import NdKkfShape


@pytest.fixture
def shape() -> NdKkfShape:
    """."""
    return NdKkfShape([2, 1], [2, 1], [1, 1])


@pytest.fixture
def gen_xz(shape: NdKkfShape) -> NdKkfMatGenXz:
    """."""
    return shape.get_mat_gen_xz()


@pytest.fixture
def gen_kf(gen_xz: NdKkfMatGenXz) -> NdKkfGenKf:
    """."""
    nx = gen_xz.gen_x.num_x
    return NdKkfGenKf(np.zeros((nx, nx)))


@pytest.fixture
def kf(gen_kf: NdKkfGenKf, gen_xz: NdKkfMatGenXz) -> KalmanFilter:
    """."""
    vec_z = np.linspace(1.0, 3.0, num=3)
    cov_zz = 3.0 * np.eye(3)
    ini_der_vars = 100 * np.linspace(1.0, 3.0, num=3)
    return gen_kf.get_kf(gen_xz, vec_z, cov_zz, ini_der_vars)


@pytest.fixture
def kf_predicted(
    kf: KalmanFilter, gen_xz: NdKkfMatGenXz, shape: NdKkfShape, gen_kf: NdKkfGenKf
) -> KalmanFilter:
    """."""
    pre = shape.get_precursors()
    pre.compute(3.0)
    gen_x = gen_xz.gen_x
    gen_x.fill_f_mat(pre, gen_kf.f_mat)
    kf.statePost[:, 0] = np.linspace(1.0, 5.0, num=5)
    factors = 6 * np.linspace(1.0, 3.0, num=3)
    gen_x.fill_q_mat(pre, factors, kf.processNoiseCov)
    kf.predict(None)
    return kf



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_vel.fundamental import FundMatCv
from kinematic_tracker.matrices.const_vel.proc_cov import WhiteNoiseMatCv


def test_proc_cov_cv() -> None:
    """."""
    wn_mat = WhiteNoiseMatCv()
    wn_mat.compute(6.0)
    assert wn_mat.q_mat == pytest.approx(np.array(((72.0, 18.0), (18.0, 6.0))))


def test_proc_cov_add_cv() -> None:
    """."""
    wn_mat = WhiteNoiseMatCv()
    wn_mat.compute(0.1)
    q1 = wn_mat.q_mat.copy()
    wn_mat.compute(0.2)
    q2 = wn_mat.q_mat.copy()
    wn_mat.compute(0.3)
    q3 = wn_mat.q_mat.copy()
    fu_mat = FundMatCv()
    fu_mat.compute(0.2)
    f_mat = fu_mat.f_mat
    diff = np.dot(f_mat, np.dot(q1, f_mat.T)) + q2 - q3
    assert diff == pytest.approx(0.0)



"""."""

import pytest

from kinematic_tracker.nd.precursors import NdKkfPrecursors


def test_nd_kkf_pre_init_simple(pre: NdKkfPrecursors) -> None:
    """."""
    assert pre.kin_mat_gen[0] is None
    for order in (1, 2, 3, 4):
        assert pre.kin_mat_gen[order].num_x == order
    assert pre.orders_set == {2, 3}


def test_nd_kkf_pre_init_asserts() -> None:
    """."""
    with pytest.raises(AssertionError):
        NdKkfPrecursors([])
    with pytest.raises(AssertionError):
        NdKkfPrecursors([0, 1])
    with pytest.raises(AssertionError):
        NdKkfPrecursors([5, 1])



"""."""

import numpy as np

from .det_type import MT


def inv_out(mat: MT, out: MT) -> None:
    try:
        out[:] = np.linalg.inv(mat)
    except np.linalg.LinAlgError:
        print('\n', mat)
        raise



"""."""

import numpy as np
import pytest

from cv2 import KalmanFilter

from kinematic_tracker.nd.gen_kf import NdKkfGenKf
from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz
from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.nd.shape import NdKkfShape
from kinematic_tracker.proc_noise.meta import ProcNoiseMeta
from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd
from kinematic_tracker.tracker.kkf import NdKkf


@pytest.fixture
def pre() -> NdKkfPrecursors:
    return NdKkfPrecursors([2, 1])


@pytest.fixture
def gen_xz(shape: NdKkfShape) -> NdKkfMatGenXz:
    return shape.get_mat_gen_xz()


@pytest.fixture
def shape() -> NdKkfShape:
    """."""
    return NdKkfShape([2, 1], [3, 2], [1, 1])


@pytest.fixture
def kf(gen_xz: NdKkfMatGenXz) -> KalmanFilter:
    """."""
    vec_z = np.linspace(1.0, 5.0, num=5)
    cov_zz = 6.0 * np.eye(5)
    ini_der_vars = np.linspace(31.0, 35.0, num=5)
    gen_kf = NdKkfGenKf(np.eye(8))
    return gen_kf.get_kf(gen_xz, vec_z, cov_zz, ini_der_vars)


@pytest.fixture
def kkf_wn(kf: KalmanFilter, gen_xz: NdKkfMatGenXz, pn_meta_fd: ProcNoiseMeta) -> NdKkf:
    """."""
    wn_fd = NdKkfWhiteNoiseFd(gen_xz.gen_x, pn_meta_fd, kf.statePre[:, 0])
    return NdKkf(kf, wn_fd)



"""Check the magic method __deepcopy__."""

import copy

from kinematic_tracker.tracker.kkf import NdKkf


def test_nd_kkf_deepcopy(kkf_wn: NdKkf) -> None:
    """."""
    assert kkf_wn.kalman_filter.gain.shape == (8, 5)
    kkf_cp = copy.deepcopy(kkf_wn)
    assert kkf_cp.kalman_filter.measurementMatrix.shape == (5, 8)
    assert kkf_cp.kalman_filter.measurementNoiseCov.shape == (5, 5)
    assert kkf_cp.kalman_filter.gain.shape == (8, 5)



"""."""

import numpy as np
import pytest

from kinematic_tracker.association.metric_size_modulated_mahalanobis import (
    MetricSizeModulatedMahalanobis,
)


def test_normal(
    det_rz: list[np.ndarray[int, float]], driver: MetricSizeModulatedMahalanobis
) -> None:
    driver.comp_sizes_dia_cov_rs(det_rz)
    ref = [[4.0, 6.25, 9.0], [6.25, 9.0, 12.25]]
    assert driver.sizes_dia_cov_rs[:2] == pytest.approx(np.array(ref))
    assert driver.sizes_dia_cov_rs[2:] == pytest.approx(0.0)


def test_empty_det(driver: MetricSizeModulatedMahalanobis) -> None:
    driver.comp_sizes_dia_cov_rs([])
    assert driver.sizes_dia_cov_rs == pytest.approx(0.0)



"""."""

import numpy as np

from kinematic_tracker.nd.gen_x import NdKkfMatGenX

from .meta import ProcNoiseKind, ProcNoiseMeta
from .nd_kkf_dia_noise import NdKkfDiaNoise
from .nd_kkf_white_noise_const import NdKkfWhiteNoiseConst
from .nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


TT_PN = NdKkfWhiteNoiseFd | NdKkfDiaNoise | NdKkfWhiteNoiseConst


def get_proc_noise(gen_x: NdKkfMatGenX, pn_meta: ProcNoiseMeta, ini_vec_x: np.ndarray) -> TT_PN:
    """."""
    kind = pn_meta.noise_kind
    if kind == ProcNoiseKind.CONST:
        pn = NdKkfWhiteNoiseConst(gen_x, pn_meta.factor)
    elif kind == ProcNoiseKind.FINITE_DIFF:
        pn = NdKkfWhiteNoiseFd(gen_x, pn_meta, ini_vec_x)
    elif kind == ProcNoiseKind.DIAGONAL:
        pn = NdKkfDiaNoise(pn_meta.factor)
    else:
        raise ValueError
    return pn



"""."""

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_promote_clutter(tracker: NdKkfTracker) -> None:
    """."""
    tracker.promote_clutter_eventually()
    assert len(tracker.tracks) == 0



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.proc_noise.nd_kkf_white_noise_const import NdKkfWhiteNoiseConst


@pytest.fixture
def pre() -> NdKkfPrecursors:
    return NdKkfPrecursors([2, 1])


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    return NdKkfMatGenX([2, 1], [3, 2])


@pytest.fixture
def wn_const(gen_x: NdKkfMatGenX) -> NdKkfWhiteNoiseConst:
    """."""
    return NdKkfWhiteNoiseConst(gen_x, 5.678)



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_vel.fundamental import FundMatCv


def test_get_fund_mat_cv() -> None:
    """."""
    fu_mat = FundMatCv()
    fu_mat.compute(0.1)
    assert fu_mat.f_mat == pytest.approx(np.array([[1.0, 0.1], [0.0, 1.0]]))



"""."""

from cv2 import CV_64F, KalmanFilter

from kinematic_tracker.core.kalman_cv2_copy import get_copy


def test_kalman_cv2_copy() -> None:
    """."""
    kf = KalmanFilter(8, 5, 0, CV_64F)
    cp = get_copy(kf)
    assert cp.measurementMatrix.shape == (5, 8)
    assert cp.measurementMatrix.dtype == 'float64'
    assert cp.gain.shape == (8, 5)
    assert id(cp.statePost) != id(kf.statePost)



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.proc_noise.meta import ProcNoiseMeta
from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


@pytest.fixture
def pre() -> NdKkfPrecursors:
    return NdKkfPrecursors([2, 1])


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    return NdKkfMatGenX([2, 1], [3, 2])


@pytest.fixture
def wn_fd(gen_x: NdKkfMatGenX, pn_meta_fd: ProcNoiseMeta) -> NdKkfWhiteNoiseFd:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    return NdKkfWhiteNoiseFd(gen_x, pn_meta_fd, vec_x)



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX


def test_nd_kkf_init_simple() -> None:
    """."""
    gen = NdKkfMatGenX([2, 1], [3, 2])
    assert gen.num_x == 8
    assert gen.orders_set == {1, 2}
    assert gen.num_d == 5


def test_nd_kkf_init_asserts() -> None:
    """."""
    with pytest.raises(AssertionError):
        NdKkfMatGenX([], [])
    with pytest.raises(AssertionError):
        NdKkfMatGenX([2, 1], [3, 2, 1])
    with pytest.raises(AssertionError):
        NdKkfMatGenX([0, 1], [3, 2])
    with pytest.raises(AssertionError):
        NdKkfMatGenX([5, 1], [3, 2])



"""."""

import numpy as np
import pytest

from kinematic_tracker.matrices.const_acc.fundamental import FundMatCa


def test_get_fund_mat_ca() -> None:
    """."""
    fu_gen = FundMatCa()
    fu_gen.compute(0.1)
    f_mat = fu_gen.f_mat
    ref = [[1.0, 0.1, 0.005], [0.0, 1.0, 0.1], [0.0, 0.0, 1.0]]
    assert f_mat == pytest.approx(np.array(ref))



"""."""

import numpy as np

from kinematic_tracker.proc_noise.nd_kkf_white_noise_const import NdKkfWhiteNoiseConst


def test_nd_kkf_white_noise_const_save_values(wn_const: NdKkfWhiteNoiseConst) -> None:
    """."""
    wn_const.save_values(np.empty(0))



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_size_modulated_mahalanobis import (
    MetricSizeModulatedMahalanobis,
)


@pytest.fixture
def filters() -> list[cv2.KalmanFilter]:
    """."""
    kf = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf.measurementMatrix = np.eye(6, 12)
    kf.measurementNoiseCov = np.eye(6) * 4.0
    kf.statePre = np.linspace(1.0, 12.0, num=12).reshape(12, 1)
    kf.errorCovPre = np.eye(12) * 1.0
    filters = [kf]

    kf2 = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf2.measurementMatrix = np.eye(6, 12)
    kf2.measurementNoiseCov = np.eye(6) * 4.0
    kf2.statePre = np.linspace(2.0, 13.0, num=12).reshape(12, 1)
    kf2.errorCovPre = np.eye(12) * 2.0
    filters.append(kf2)

    kf3 = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf3.measurementMatrix = np.eye(6, 12)
    kf3.measurementNoiseCov = np.eye(6) * 4.0
    kf3.statePre = np.linspace(3.0, 14.0, num=12).reshape(12, 1)
    kf3.errorCovPre = np.eye(12) * 3.0
    filters.append(kf3)

    return filters


@pytest.fixture
def driver() -> MetricSizeModulatedMahalanobis:
    return MetricSizeModulatedMahalanobis(100, 500, 1.0, 12, 6, (0, 1, 2, -3, -2, -1))



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    return NdKkfMatGenX([2, 3], [4, 5])


@pytest.fixture
def gen_xz(gen_x: NdKkfMatGenX) -> NdKkfMatGenXz:
    return NdKkfMatGenXz(gen_x, [1, 2])



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


@pytest.fixture
def tracker() -> NdKkfTracker:
    """."""
    tracker = NdKkfTracker([3, 1], [3, 3])
    tracker.association.threshold = 0.56
    tracker.set_association_metric('mahalanobis', 0.22 * 6)
    tracker.set_association_method('greedy')
    tracker.num_det_min = 2
    return tracker


@pytest.fixture
def tracker1(tracker: NdKkfTracker) -> NdKkfTracker:
    """."""
    tracker.set_measurement_cov(0.01 * np.eye(6))
    tracker.association.threshold = 0.56
    r2z = [np.linspace(1.0, 6.0, num=6)]
    r2id = [123]
    tracker.advance(1234_000_000, r2z, r2id)
    return tracker



"""."""

import pytest

from kinematic_tracker.proc_noise.meta import ProcNoiseKind, ProcNoiseMeta, get_proc_noise_meta


@pytest.fixture
def pn_meta_fd() -> ProcNoiseMeta:
    return get_proc_noise_meta(ProcNoiseKind.FINITE_DIFF, 1.0, 0.75, True, True)



"""."""

import numpy as np
import pytest

from kinematic_tracker.nd.precursors import NdKkfPrecursors


def test_nd_kkf_pre_compute(pre: NdKkfPrecursors) -> None:
    """."""
    pre.compute(3.0)
    assert pre.kin_mat_gen[1].q_mat_dt == pytest.approx(np.zeros((1, 1)))
    assert pre.kin_mat_gen[1].f_mat_dt == pytest.approx(np.eye(1))
    ref_f2 = [[1.0, 3.0], [0.0, 1.0]]
    ref_f3 = [[1.0, 3.0, 4.5], [0.0, 1.0, 3.0], [0.0, 0.0, 1.0]]
    assert pre.kin_mat_gen[2].f_mat_dt == pytest.approx(np.array(ref_f2))
    assert pre.kin_mat_gen[3].f_mat_dt == pytest.approx(np.array(ref_f3))
    ref_q2 = [[9.0, 4.5], [4.5, 3.0]]
    ref_q3 = [[12.149999999999999, 10.125, 4.5], [10.125, 9.0, 4.5], [4.5, 4.5, 3.0]]
    assert pre.kin_mat_gen[2].q_mat_dt == pytest.approx(np.array(ref_q2))
    assert pre.kin_mat_gen[3].q_mat_dt == pytest.approx(np.array(ref_q3))
    assert pre.kin_mat_gen[4].q_mat_dt == pytest.approx(np.zeros((4, 4)))
    assert pre.kin_mat_gen[4].f_mat_dt == pytest.approx(np.eye(4))



"""
This module defines the `MetricSizeModulatedMahalanobis` class.
The class extends the `MetricDriverBase` class to compute a size-modulated
Mahalanobis-distance similarity score.
"""

from typing import Sequence

import cv2
import numpy as np

from .det_type import DT, MT
from .inv_out import inv_out
from .metric_driver_base import MetricDriverBase


class MetricSizeModulatedMahalanobis(MetricDriverBase):
    """
    A class for computing a size-modulated Mahalanobis distance-based metric for association
    of detection reports and targets represented by Kalman filters.
    It extends the `MetricDriverBase` class.

    Attributes:
        mah_pre_factor: pre-factor for scaling the Mahalanobis distance.
        tmp_zx: temporary rectangular matrix for intermediate calculations.
        cov_zz: covariance matrix for the sum of the covariances.
        inv_zz: inverse of the covariance matrix.
        num_z: dimensionality of the measurement space (number of detection variables).
        ind_pos_size: location of variables in the measurement vector.
        det_rz: buffer for detections
    """

    def __init__(
        self,
        num_reports_max: int,
        num_targets_max: int,
        mah_pre_factor: float,
        num_x: int,
        num_z: int,
        ind_pos_size: Sequence[int],
    ) -> None:
        """
        Initializes the MetricSizeModulatedMahalanobis instance with the given parameters.

        Args:
            num_reports_max: maximum number of reports.
            num_targets_max: maximum number of targets.
            mah_pre_factor: pre-factor for scaling the (square) of Mahalanobis distance.
            num_x: dimensionality of the state space used for allocating auxiliary buffers.
            num_z: dimensionality of the measurement space used for allocating auxiliary buffers.
            ind_pos_size: Location of variables in the measurement vector.
                          By default, the indices will be 0,1,2,-3,-2,-1, i.e.
                          the center of the cuboid is taken as first three variables
                          of the measurement vector and sizes (dimensions) of the
                          cuboids are taken from the last three variables.

        Raises:
            AssertionError: If `num_x` is less than `num_z`.
        """
        assert num_x >= num_z, f'Mixed up x and z? num_x = {num_x} < {num_z} = num_z'
        super().__init__(num_reports_max, num_targets_max, num_z)
        self.ind_pos_size = np.array(ind_pos_size, dtype=int)
        assert self.ind_pos_size.ndim == 1, 'Sequence of should be one-dimensional'
        assert self.ind_pos_size.size % 2 == 0, 'Number of positions and sizes should be the same'
        assert self.ind_pos_size.size <= num_z, 'Number variables should verify inequality'
        num_sizes = self.ind_pos_size.size // 2
        self.pos_ind = self.ind_pos_size[:num_sizes]
        self.size_ind = self.ind_pos_size[num_sizes:]
        self.mah_pre_factor = mah_pre_factor
        self.tmp_zx = np.zeros((num_z, num_x))
        self.cov_zz = np.zeros((num_z, num_z))
        self.inv_zz = np.zeros((num_z, num_z))
        self.num_z = num_z
        self.det_rz = np.zeros((self.num_reports_max, num_z))
        self.sizes_dia_cov_rs = np.zeros((self.num_reports_max, num_sizes))
        self.sizes_dia_cov_s = np.zeros(num_sizes)
        self.size_indices = (self.size_ind, self.size_ind)
        self.pos_indices = (self.pos_ind, self.pos_ind)

    def comp_sizes_dia_cov_rs(self, det_rz: DT) -> None:
        num_r = len(det_rz)
        if num_r > 0:
            np.copyto(self.det_rz[:num_r], det_rz)
            np.take(self.det_rz[:num_r], self.size_ind, axis=1, out=self.sizes_dia_cov_rs[:num_r])
            np.square(self.sizes_dia_cov_rs[:num_r], out=self.sizes_dia_cov_rs[:num_r])
            np.divide(self.sizes_dia_cov_rs[:num_r], 4, out=self.sizes_dia_cov_rs[:num_r])

    def compute_metric(self, det_rz: DT, filters: Sequence[cv2.KalmanFilter]) -> MT:
        """
        Computes the Mahalanobis distance-based metric for associating detections
        with Kalman filters.

        Args:
            det_rz: the input sequence of detection reports (measurements).
            filters: the input sequence of Kalman filters representing the tracked targets.

        Returns:
            np.ndarray[tuple[int, int], float]: A 2D array containing the computed metrics.
            The array has shape (num_reports, num_targets) and located at the beginning
            of the (larger) `self.metric_rt` buffer.

        Raises:
            np.linalg.LinAlgError: If the covariance matrix is singular and cannot be inverted.
        """
        self.comp_sizes_dia_cov_rs(det_rz)

        rect_chunk = self.get_rect_chunk(len(det_rz), len(filters))
        for t, kf in enumerate(filters):
            np.dot(kf.measurementMatrix, kf.statePre, out=self.vec_z)
            np.square(self.vec_z[self.size_ind, 0], out=self.sizes_dia_cov_s)  # xxx tmp buf
            np.divide(self.sizes_dia_cov_s, 4, out=self.sizes_dia_cov_s)
            np.dot(kf.measurementMatrix, kf.errorCovPre, out=self.tmp_zx)
            np.dot(self.tmp_zx, kf.measurementMatrix.T, out=self.cov_zz)
            np.add.at(self.cov_zz, self.size_indices, self.sizes_dia_cov_s)
            np.add.at(self.cov_zz, self.pos_indices, self.sizes_dia_cov_s)
            np.add(self.cov_zz, kf.measurementNoiseCov, out=self.cov_zz)
            for r, det_z in enumerate(det_rz):
                np.add.at(self.cov_zz, self.size_indices, self.sizes_dia_cov_rs[r])
                np.add.at(self.cov_zz, self.pos_indices, self.sizes_dia_cov_rs[r])
                inv_out(self.cov_zz, self.inv_zz)
                sqr_distance = cv2.Mahalanobis(self.vec_z, det_z, self.inv_zz) ** 2
                rect_chunk[r, t] = sqr_distance
        rect_chunk *= self.mah_pre_factor / (2.0 * self.num_z)
        rect_chunk[rect_chunk > 150.0] = 150
        np.exp(-rect_chunk, out=rect_chunk)
        return rect_chunk



"""."""

from kinematic_tracker.nd.gen_xz import NdKkfMatGenXz
from kinematic_tracker.nd.precursors import NdKkfPrecursors
from kinematic_tracker.nd.shape import NdKkfShape


def test_nd_kkf_get_precursors(shape: NdKkfShape) -> None:
    """."""
    pre = shape.get_precursors()
    assert isinstance(pre, NdKkfPrecursors)


def test_nd_kkf_get_gen_xz(shape: NdKkfShape) -> None:
    """."""
    gen_xz = shape.get_mat_gen_xz()
    assert isinstance(gen_xz, NdKkfMatGenXz)



"""Module providing utility functions for tracking IDs."""


def get_free_id(ids: list[int]) -> int:
    """
    Find the smallest non-negative integer that is not in the given list of IDs.

    This function results in the locally unique IDs. However, will reuse IDs
    if they are not currently in use.

    Args:
        ids (list[int]): A list of integers representing used IDs.

    Returns:
        int: The smallest non-negative integer not present in the list.
    """
    track_id = 0
    for track_id in range(len(ids) + 1):
        if track_id not in ids:
            break
    return track_id



"""."""

import numpy as np

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.proc_noise.kind import ProcNoiseKind
from kinematic_tracker.proc_noise.meta import get_proc_noise_meta
from kinematic_tracker.proc_noise.nd_kkf_proc_noise_factory import get_proc_noise
from kinematic_tracker.proc_noise.nd_kkf_white_noise_const import NdKkfWhiteNoiseConst


def test_get_proc_noise_white_noise_const(gen_x: NdKkfMatGenX) -> None:
    """."""
    vec_x = np.linspace(1.0, 8.0, num=8)
    pn_meta_const = get_proc_noise_meta(ProcNoiseKind.CONST, 0.5, 0.5, True, True)
    pn = get_proc_noise(gen_x, pn_meta_const, vec_x)
    assert isinstance(pn, NdKkfWhiteNoiseConst)



"""."""

import pytest

from kinematic_tracker.nd.derivative_sorted import DerivativeSorted
from kinematic_tracker.nd.gen_x import NdKkfMatGenX


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    return NdKkfMatGenX([3, 1], [2, 2])


@pytest.fixture
def der_sorted(gen_x: NdKkfMatGenX) -> DerivativeSorted:
    """."""
    return DerivativeSorted(gen_x)



"""."""


class CreationId:
    """."""

    def __init__(self) -> None:
        """."""
        self.next_id: int = -1

    def get_next_id(self) -> int:
        """."""
        self.next_id += 1
        return self.next_id



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


@pytest.fixture
def filters() -> list[cv2.KalmanFilter]:
    """."""
    kf = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf.measurementMatrix = np.eye(6, 12)
    kf.statePre = np.linspace(1.0, 12.0, num=12).reshape(12, 1)
    filters = [kf]

    kf2 = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf2.measurementMatrix = np.eye(6, 12)
    kf2.statePre = np.linspace(2.0, 13.0, num=12).reshape(12, 1)
    filters.append(kf2)

    kf3 = cv2.KalmanFilter(12, 6, 0, cv2.CV_64F)
    kf3.measurementMatrix = np.eye(6, 12)
    kf3.statePre = np.linspace(3.0, 14.0, num=12).reshape(12, 1)
    filters.append(kf3)

    return filters


@pytest.fixture
def driver() -> MetricGIoUAligned:
    return MetricGIoUAligned(100, 500, 6, (0, 1, 2, 3, 4, 5))



"""."""

from kinematic_tracker.core.track_free_id import get_free_id


class MockTrack:
    def __init__(self, self_id: int) -> None:
        """."""
        self.id = self_id


def test_get_free_id() -> None:
    """."""
    tracks = [MockTrack(0), MockTrack(1)]
    tracks[0].id = 0
    tracks[1].id = 1
    assert get_free_id(tracks) == 2



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


@pytest.fixture
def filters() -> list[cv2.KalmanFilter]:
    """."""
    kf = cv2.KalmanFilter(16, 10, 0, cv2.CV_64F)
    kf.measurementMatrix = np.eye(10, 16)
    kf.statePre = np.linspace(1.0, 16.0, num=16).reshape(16, 1)
    vec_x = np.array(
        (1.0, 2.0, 3.0, 0.0, 0.0, 0.0, 0.0, 4.0, 5.0, 6.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0)
    ).reshape(16, 1)
    kf.statePre = vec_x
    filters = [kf]

    kf2 = cv2.KalmanFilter(16, 10, 0, cv2.CV_64F)
    kf2.measurementMatrix = np.eye(10, 16)
    vec_x = np.array(
        (2.0, 3.0, 4.0, 0.0, 0.0, 0.0, 0.0, 5.0, 6.0, 7.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0)
    ).reshape(16, 1)
    kf2.statePre = vec_x
    filters.append(kf2)

    kf3 = cv2.KalmanFilter(16, 10, 0, cv2.CV_64F)
    kf3.measurementMatrix = np.eye(10, 16)
    vec_x = np.array(
        (3.0, 4.0, 5.0, 0.0, 0.0, 0.0, 0.0, 6.0, 7.0, 8.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0)
    ).reshape(16, 1)
    kf3.statePre = vec_x
    filters.append(kf3)

    return filters


@pytest.fixture
def det_rz() -> list[np.ndarray]:
    """."""
    vec_z1 = np.array((1.0, 2.0, 3.0, 0.0, 0.0, 0.0, 0.0, 4.0, 5.0, 6.0))
    vec_z2 = np.array((2.0, 3.0, 4.0, 0.0, 0.0, 0.0, 0.0, 5.0, 6.0, 7.0))
    return [vec_z1, vec_z2]


def test_at_end(det_rz: list[np.ndarray[int, float]], filters: list[cv2.KalmanFilter]) -> None:
    """."""
    driver = MetricGIoUAligned(100, 500, 10, (0, 1, 2, -3, -2, -1))
    assert driver.ind_pos_size == pytest.approx((0, 1, 2, -3, -2, -1))
    assert driver.ind_pos_size.dtype == int
    metric = driver.compute_metric(det_rz, filters)
    ref = [
        [1.0, 0.6318122555410691],
        [0.6318122555410691, 1.0],
        [0.4686147186147186, 0.6735666418466121],
    ]
    assert metric == pytest.approx(np.array(ref).T)
    assert np.shares_memory(metric, driver.metric_rt)



"""."""

import numpy as np
import pytest

from kinematic_tracker.tracker.tracker import NdKkfTracker


@pytest.fixture
def tracker() -> NdKkfTracker:
    return NdKkfTracker([3, 1], [2, 2], [2, 1])


def test_set_measurement_std_dev_at_start_part0_der_order0(tracker: NdKkfTracker) -> None:
    assert tracker.cov_zz is None
    tracker.set_measurement_std_dev(0, 0.5, 0)
    assert isinstance(tracker.cov_zz, np.ndarray)
    assert tracker.cov_zz.shape == (6, 6)
    for i in (1, 3, 4, 5):
        assert np.isnan(tracker.cov_zz[i, i])
        tracker.cov_zz[i, i] = (i + 1) * 10

    ref_zz = [
        [0.25, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 20.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.25, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 40.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 50.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 60.0],
    ]
    assert tracker.cov_zz == pytest.approx(np.array(ref_zz))


def test_set_measurement_std_dev_at_start_part0_der_order1(tracker: NdKkfTracker) -> None:
    tracker.set_measurement_std_dev(0, 0.5, 1)
    assert isinstance(tracker.cov_zz, np.ndarray)
    for i in (0, 2, 4, 5):
        assert np.isnan(tracker.cov_zz[i, i])
        tracker.cov_zz[i, i] = (i + 1) * 10

    ref_zz = [
        [10.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.25, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 30.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.25, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 50.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 60.0],
    ]
    assert tracker.cov_zz == pytest.approx(np.array(ref_zz))


def test_set_measurement_std_dev_at_start_part1_der_order0(tracker: NdKkfTracker) -> None:
    tracker.set_measurement_std_dev(1, 0.5, 0)
    assert isinstance(tracker.cov_zz, np.ndarray)
    for i in (0, 1, 2, 3):
        assert np.isnan(tracker.cov_zz[i, i])
        tracker.cov_zz[i, i] = (i + 1) * 10

    ref_zz = [
        [10.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 20.0, 0.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 30.0, 0.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 40.0, 0.0, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.25, 0.0],
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.25],
    ]
    assert tracker.cov_zz == pytest.approx(np.array(ref_zz))


def test_set_measurement_std_dev_at_start_part1_der_order1(tracker: NdKkfTracker) -> None:
    with pytest.raises(ValueError):
        tracker.set_measurement_std_dev(1, 0.5, 1)


def test_set_measurement_std_dev_on_the_go(tracker: NdKkfTracker) -> None:
    tracker.cov_zz = -999 * np.ones((6, 6))
    tracker.set_measurement_std_dev(1, 0.5, 0)
    ref_zz = [
        [-999.0, -999.0, -999.0, -999.0, -999.0, -999.0],
        [-999.0, -999.0, -999.0, -999.0, -999.0, -999.0],
        [-999.0, -999.0, -999.0, -999.0, -999.0, -999.0],
        [-999.0, -999.0, -999.0, -999.0, -999.0, -999.0],
        [-999.0, -999.0, -999.0, -999.0, 0.25, -999.0],
        [-999.0, -999.0, -999.0, -999.0, -999.0, 0.25],
    ]
    assert tracker.cov_zz == pytest.approx(np.array(ref_zz))



"""."""

from kinematic_tracker.proc_noise.nd_kkf_white_noise_const import NdKkfWhiteNoiseConst


def test_nd_kkf_white_noise_const_repr(wn_const: NdKkfWhiteNoiseConst) -> None:
    """."""
    assert repr(wn_const) == 'NdKkfWhiteNoiseConst(5.678)'



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_size_modulated_mahalanobis import (
    MetricSizeModulatedMahalanobis,
)


def test_compute_metric(
    det_rz: list[np.ndarray[int, float]],
    filters: list[cv2.KalmanFilter],
    driver: MetricSizeModulatedMahalanobis,
) -> None:
    metric = driver.compute_metric(det_rz, filters)
    ref = [
        [1.0, 0.9760750935048036, 0.9228298024606032],
        [0.9805168491782723, 1.0, 0.9851645434569364],
    ]
    assert metric == pytest.approx(np.array(ref))
    assert np.shares_memory(metric, driver.metric_rt)



"""The parameters for initialization of the Kalman filter and ProcNoise objects.

The class provides the corresponding methods as well.
"""

from typing import Sequence

import numpy as np

from .gen_x import NdKkfMatGenX
from .gen_xz import NdKkfMatGenXz
from .precursors import NdKkfPrecursors


class NdKkfShape:
    """."""

    def __init__(
        self, orders_x: Sequence[int], num_dims: Sequence[int], orders_z: Sequence[int]
    ) -> None:
        """The constructor

        Args:
            orders_x: point-motion models of the partial kinematic states (CP - 1, CV - 2, CA - 3, CJ - 4).
            num_dims: numbers of Cartesian degrees of freedom for every partial state.
            orders_z: description of detection content similar to the point-motion models of the states.
        """
        self.orders_x = np.array(orders_x, dtype=int)
        self.num_dims = np.array(num_dims, dtype=int)
        self.orders_z = np.array(orders_z, dtype=int)

    def __repr__(self) -> str:
        """."""
        return (
            f'NdKkfShape( '
            f'orders_x {self.orders_x} '
            f'num_dims {self.num_dims} '
            f'orders_z {self.orders_z})'
        )

    def get_precursors(self) -> NdKkfPrecursors:
        """Compose a ready-to-use precursor object.

        Returns:
            The precursor object.
        """
        return NdKkfPrecursors(self.orders_x)

    def _get_mat_gen_x(self) -> NdKkfMatGenX:
        """Compose a ready-to-use state matrix generator.

        Returns:
            The state matrix generator object.
        """
        return NdKkfMatGenX(self.orders_x, self.num_dims)

    def get_mat_gen_xz(self) -> NdKkfMatGenXz:
        """Compose a ready-to-use measurement matrix generator.

        Returns:
            The Kalman-matrices generator object.
        """
        gen_x = self._get_mat_gen_x()
        return NdKkfMatGenXz(gen_x, self.orders_z)



"""."""

import cv2
import numpy as np
import pytest

from kinematic_tracker.association.metric_giou_aligned import MetricGIoUAligned


def test_min(
    det_rz: list[np.ndarray[int, float]], filters: list[cv2.KalmanFilter], driver: MetricGIoUAligned
) -> None:
    """."""
    metric = driver.compute_metric(det_rz, filters)
    ref = [
        [1.0, 0.6318122555410691],
        [0.6318122555410691, 1.0],
        [0.4686147186147186, 0.6735666418466121],
    ]
    assert metric == pytest.approx(np.array(ref).T)
    assert np.shares_memory(metric, driver.metric_rt)



"""."""

import numpy as np

from kinematic_tracker.tracker.tracker import NdKkfTracker


def test_nd_kkf_tracker_upd_for_loose_reports1(tracker1: NdKkfTracker) -> None:
    """."""
    o2z = [0.1 + np.linspace(1.0, 6.0, num=6)]
    o2id = [124]
    tracker1.update_for_loose_reports({0: 0}, o2z, o2id)
    assert len(tracker1.tracks) == 1



"""
This module provides a utility function to determine which sequential indices
are absent in a given list.
"""

from typing import Sequence


def get_loose_indices(indices: Sequence[int], num_indices: int) -> set[int]:
    """
    Calculate the set of indices that are not included in the given sequence of indices.

    Args:
        indices: A sequence of integers representing the indices to exclude.
        num_indices: The total number of indices to consider.

    Returns:
        set[int]: A set of indices that are not in the input sequence.
    """
    return set(range(num_indices)) - set(indices)



"""."""

from kinematic_tracker.core.free_id import get_free_id


def test_get_free_id() -> None:
    """."""
    assert get_free_id([-1, -1]) == 0
    assert get_free_id([-1, 1]) == 0
    assert get_free_id([-1, 0]) == 1
    assert get_free_id([1, 2]) == 0
    assert get_free_id([0, 1]) == 2



"""."""

import pytest

from kinematic_tracker.proc_noise.nd_kkf_white_noise_fd import NdKkfWhiteNoiseFd


def test_nd_kkf_proc_noise_repr(wn_fd: NdKkfWhiteNoiseFd) -> None:
    """."""
    ref = """NdKkfWhiteNoiseFd(ProcNoiseMeta(ProcNoiseKind.FINITE_DIFF
    factor 1.0
    mix_weight 0.75
    mult_by_last_derivative True
    is_dynamic_mixing True))"""
    assert repr(wn_fd) == ref
    assert wn_fd.der_mixer.values == pytest.approx([2, 4, 6, 7, 8])



"""."""

from kinematic_tracker.core.creation_id import CreationId


def test_creation_id() -> None:
    """."""
    driver = CreationId()
    assert driver.get_next_id() == 0
    assert driver.get_next_id() == 1
    assert driver.get_next_id() == 2



"""."""

import pytest

from kinematic_tracker.nd.gen_x import NdKkfMatGenX
from kinematic_tracker.nd.precursors import NdKkfPrecursors


@pytest.fixture
def pre() -> NdKkfPrecursors:
    return NdKkfPrecursors([2, 1])


@pytest.fixture
def gen_x() -> NdKkfMatGenX:
    return NdKkfMatGenX([2, 1], [3, 2])
